#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 15:09
from dotenv import load_dotenv

"""
优先级规则（从高到低）：
1. 代码中传入的参数（如 Settings(headless=False)）
2. 环境变量
3. .env 文件
4. 类属性默认值
"""
load_dotenv(override=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 17:47
import asyncio
import io
import time
import traceback
from abc import ABC, abstractmethod
from functools import wraps
from pathlib import Path
from typing import IO, Optional, cast, TypeAlias, Union

from httpx import AsyncClient
from loguru import logger
# noinspection PyProtectedMember
from loguru._logger import context as logger_context
from playwright.async_api import TimeoutError
from pydantic import TypeAdapter
from pydantic_ai import ModelRetry, RunContext, Agent

from .config import global_settings
from .deps import AgentDeps, ToolParams, ToolResult, StepInfo, LocationToolParams, ClickToolParams, \
    InputToolParams, SwipeToolParams, SwipeFromCoordinateToolParams, OpenUrlToolParams, ScreenInfo, AgentContext, \
    WaitToolParams, AssertContainsParams, MarkFailedParams, AssertNotContainsParams, ToolResultWithOutput
from .device import AndroidDevice, WebDevice
from .util.adb_tool import AdbDeviceProxy
from .util.js_tool import JSTool
from .util.platform import get_client_url_schema

storage_client = global_settings.storage_client

AgentDepsType: TypeAlias = AgentDeps[
    Union[WebDevice, AndroidDevice],
    Union['WebAgentTool', 'AndroidAgentTool'],
]


class ToolHandler:
    """工具调用处理类，用于处理工具函数的前置和后置操作"""

    def __init__(self, *args, **kwargs):
        self.ctx: Optional[RunContext[AgentDepsType]] = None
        self.step_params: Optional[ToolParams] = None
        self.step_info: Optional[StepInfo] = None

        for arg in [*args, *kwargs.values()]:
            if isinstance(arg, RunContext):
                self.ctx = cast(RunContext[AgentDepsType], arg)
                continue
            if isinstance(arg, ToolParams):
                self.step_params = arg
                continue

    @property
    def context(self) -> AgentContext:
        return self.ctx.deps.context

    @property
    def current_step(self) -> StepInfo:
        return self.ctx.deps.context.current_step

    async def pre_handle(self, func):
        """工具的前置处理"""
        if not all([self.ctx, self.step_params]):
            return
        if self.ctx.deps.context.current_step.parallel_tool_calls:
            raise ModelRetry('only use one tool at a time')

        self.current_step.params = self.step_params.model_dump(
            exclude_defaults=True,
            exclude_none=True
        )
        self.current_step.action = self.current_step.params.pop('action')

        if self.ctx.deps.settings.debug and isinstance(self.step_params, LocationToolParams):
            if isinstance(self.ctx.deps.device, WebDevice):
                await JSTool.add_highlight_element(self.ctx.deps.device.page, self.step_params.element_bbox)

    async def post_handle(self, tool_result: ToolResult):
        """工具的后置处理"""
        if not all([self.ctx, self.step_params]):
            return
        self.current_step.is_success = tool_result.is_success


def limit_recursion(max_depth):
    """用来限制工具内部递归次数"""
    def decorator(func):
        depth = 0

        @wraps(func)
        async def wrapper(*args, **kwargs):
            nonlocal depth
            if depth >= max_depth:
                return ToolResult.failed()
            depth += 1
            try:
                return await func(*args, **kwargs)
            finally:
                depth -= 1

        return wrapper

    return decorator


def tool(f=None, *, after_delay=0, before_delay=0):
    """
    工具函数装饰器，用于标记函数为工具函数，并自动记录步骤信息
    :param f: 被装饰的函数
    :param before_delay: 操作前的等待时间，单位为秒，默认为0
    :param after_delay: 操作后的等待时间，单位为秒，默认为0
    :return: 装饰后的函数
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            tool_handler = ToolHandler(*args, **kwargs)
            await tool_handler.pre_handle(func)

            try:
                # 工具执行
                await asyncio.sleep(before_delay)
                result = await func(*args, **kwargs)
                await asyncio.sleep(after_delay)  # 避免页面渲染慢，不稳定
                await tool_handler.post_handle(result)
                return result
            except Exception as e:
                logger.error(f"Error occurred in tool '{func.__name__}': {str(e)}")
                format_exc = traceback.format_exc()
                for line in format_exc.splitlines():
                    logger.error(line)

                raise ModelRetry(f"Error occurred, try call '{func.__name__}' again")

        wrapper.is_tool = True
        return wrapper

    if f is not None:
        return decorator(f)
    return decorator


class AgentTool(ABC):
    OMNI_BASE_URL = global_settings.omni_base_url
    OMNI_KEY = global_settings.omni_key

    @property
    def tools(self) -> list:
        result = []
        for item in dir(self):
            if item.startswith('_') or item in ['tools', 'tear_down']:
                continue
            value = getattr(self, item)
            if callable(value) and hasattr(value, 'is_tool'):
                result.append(value)

        return result

    @staticmethod
    async def _upload_cos(file: IO[bytes], prefix='page-eyes-agent/', suffix='.png') -> str:
        return await storage_client.async_upload_file(file, prefix=prefix, suffix=suffix)

    async def _parse_element(self, file: Optional[IO[bytes]] = None, image_url: Optional[str] = None):
        url = f'{self.OMNI_BASE_URL}/omni/parse/'
        if not file and not image_url:
            raise ValueError('请提供file或image_url')
        trace_id = logger_context.get().get('trace_id')
        headers = {'X-Trace-Id': trace_id} if trace_id else None
        async with AsyncClient(timeout=300, headers=headers) as client:
            response = await client.post(url, files={'file': file}, data={'key': self.OMNI_KEY})
            response.raise_for_status()
            return response.json()

    @staticmethod
    @abstractmethod
    async def screenshot(ctx: RunContext[AgentDepsType]) -> io.BytesIO:
        raise NotImplementedError

    async def get_screen(self, ctx: RunContext[AgentDepsType], parse_element: bool = True) -> ScreenInfo:
        image_buffer = await self.screenshot(ctx)
        if parse_element:
            parsed_data = await self._parse_element(image_buffer)
            image_url = parsed_data.get('labeled_image_url') or ''
            parsed_content_list = parsed_data.get('parsed_content_list') or []
            logger.info(f'👁‍🗨 Get screen element：{image_url}')
            if not parsed_content_list:
                raise Exception(f'Screen parsed error! {parsed_data}')
        else:
            image_url = await self._upload_cos(image_buffer, suffix=Path(image_buffer.name).suffix)
            parsed_content_list = []
            logger.info(f'👁‍🗨 Get screen url：{image_url}')

        # 将当前屏幕信息记录到上下文
        ctx.deps.context.current_step.image_url = image_url
        ctx.deps.context.current_step.screen_elements = parsed_content_list
        # 仅保留必要的字段给LLM
        parsed_elements = TypeAdapter(list[dict]).dump_python(
            parsed_content_list,
            exclude={'__all__': {'type', 'interactivity', 'source'}}
        )
        return ScreenInfo(image_url=image_url, screen_elements=parsed_elements)

    @abstractmethod
    async def get_screen_info(self, ctx: RunContext[AgentDepsType]) -> ToolResultWithOutput[dict]:
        raise NotImplementedError

    @abstractmethod
    async def open_url(self, ctx: RunContext[AgentDepsType], params: OpenUrlToolParams) -> ToolResult:
        raise NotImplementedError

    @abstractmethod
    async def click(self, ctx: RunContext[AgentDepsType], params: ClickToolParams) -> ToolResult:
        raise NotImplementedError

    @abstractmethod
    async def input(self, ctx: RunContext[AgentDepsType], params: InputToolParams) -> ToolResult:
        raise NotImplementedError

    @abstractmethod
    async def swipe(self, ctx: RunContext[AgentDepsType], params: SwipeToolParams) -> ToolResult:
        raise NotImplementedError

    @abstractmethod
    async def tear_down(self, ctx: RunContext[AgentDepsType], params: ToolParams) -> ToolResult:
        raise NotImplementedError

    @tool(after_delay=0)
    async def wait(self, ctx: RunContext[AgentDepsType], params: WaitToolParams) -> ToolResult:
        """
        在任务中等待或停留指定的超时时间（timeout），单位：秒，等待过程中可期望指定的关键字出现
        """
        if params.expect_keywords is None:
            logger.info(f'Wait for timeout {params.timeout}s')
            await asyncio.sleep(params.timeout)
            return ToolResult.success()
        else:
            logger.info(f'Wait up to {params.timeout}s '
                        f'for the keywords {params.expect_keywords} to appear on the screen.')
            st = time.time()
            while time.time() - st < params.timeout:
                result = await self.expect_screen_contains(ctx, params.expect_keywords)
                if result.is_success:
                    return result
                await asyncio.sleep(2)
            else:
                return ToolResult.failed()
    
    async def _parse_screen_keywords(self, ctx: RunContext[AgentDepsType], keywords: list[str]) -> tuple[list, list]:
        screen_info: ScreenInfo = await self.get_screen(ctx, parse_element=True)
        elements_str = str(screen_info.screen_elements)
        contains, not_contains = [], []
        for keyword in keywords:
            if keyword in elements_str:
                contains.append(keyword)
            else:
                not_contains.append(keyword)
        return contains, not_contains

    async def expect_screen_contains(
            self,
            ctx: RunContext[AgentDepsType],
            keywords: list[str]
    ) -> ToolResult:
        contains, not_contains = await self._parse_screen_keywords(ctx, keywords)
        if len(not_contains) > 0:
            logger.warning(f'Screen does not contain expected keywords:"{not_contains}"')
            return ToolResult.failed()
        else:
            return ToolResult.success()

    async def expect_screen_not_contains(
            self,
            ctx: RunContext[AgentDepsType],
            keywords: list[str]
    ) -> ToolResult:
        contains, not_contains = await self._parse_screen_keywords(ctx, keywords)
        if len(contains) > 0:
            logger.warning(f'Screen contains unexpected keywords:"{contains}"')
            return ToolResult.failed()
        else:
            return ToolResult.success()

    @tool(before_delay=2)
    async def assert_screen_contains(
            self,
            ctx: RunContext[AgentDepsType],
            params: AssertContainsParams
    ) -> ToolResult:
        """
        检查屏幕中是否出现或包含指定的多个关键字内容，如果是则 is_success=True, 否则 is_success=False
        """
        return await self.expect_screen_contains(ctx, params.expect_keywords)

    @tool(before_delay=2)
    async def assert_screen_not_contains(
            self,
            ctx: RunContext[AgentDepsType],
            params: AssertNotContainsParams
    ) -> ToolResult:
        """
        检查屏幕中是否不出现或不包含指定的多个关键字内容，如果是则 is_success=True, 否则 is_success=False
        """
        return await self.expect_screen_not_contains(ctx, params.unexpect_keywords)

    @tool
    async def mark_failed(
            self,
            ctx: RunContext[AgentDepsType],
            params: MarkFailedParams
    ) -> ToolResult:
        """
        Mark the task as failed and terminate immediately if an element is not found or is not actionable.
        """
        logger.info(f'Mark task failed, reason: {params.reason}')
        ctx.deps.context.set_step_failed(params.reason)
        return ToolResult.success()


class WebAgentTool(AgentTool):

    @staticmethod
    async def screenshot(ctx: RunContext[AgentDepsType]) -> io.BytesIO:
        screenshot = await ctx.deps.device.page.screenshot(full_page=False, style='#option-el-box {display: none;}')
        image_buffer = io.BytesIO(screenshot)
        image_buffer.name = 'screen.png'
        return image_buffer

    @tool
    async def get_screen_info(self, ctx: RunContext[AgentDepsType]) -> ToolResultWithOutput[dict]:
        """
        获取当前屏幕信息，screen_elements 包含所有解析到的元素信息，列表顺序即为屏幕元素的排列顺序，从左到右，从上到下
        每个元素包含以下字段：
        id: 元素的id
        bbox: 元素的相对坐标，格式为 (x1, y1, x2, y2)
        content: 元素描述信息
        left_elem_ids: 该元素左侧的元素列表
        right_elem_ids: 该元素右侧的元素列表
        top_elem_ids: 该元素上方的元素列表
        bottom_elem_ids: 该元素下方的元素列表
        """
        screen_info = await self.get_screen(ctx)
        return ToolResultWithOutput.success(screen_info.model_dump(include={'screen_elements'}))

    @tool
    async def tear_down(self, ctx: RunContext[AgentDepsType], params: ToolParams) -> ToolResult:
        """
        任务完成或结束后的清理操作
        """
        await JSTool.remove_highlight_element(ctx.deps.device.page)
        await self.get_screen(ctx, parse_element=False)

        if ctx.deps.device.playwright is not None:
            await ctx.deps.device.context.close()
            await ctx.deps.device.playwright.stop()
        return ToolResult.success()

    @tool(after_delay=2)
    async def open_url(self, ctx: RunContext[AgentDepsType], params: OpenUrlToolParams) -> ToolResult:
        """
        使用设备打开URL
        """
        await ctx.deps.device.page.goto(params.url, wait_until='networkidle')
        return ToolResult.success()

    @tool(after_delay=2)
    async def click(self, ctx: RunContext[AgentDepsType], params: ClickToolParams) -> ToolResult:
        """
        点击设备屏幕指定的元素, element_bbox 不能为空
        """
        x, y = params.get_coordinate(ctx.deps.device.device_size, params.position, params.offset)
        logger.info(f'click coordinate ({x}, {y})')
        await JSTool.add_highlight_position(ctx.deps.device.page, x, y)
        try:
            if params.file_path:
                logger.info(f'upload file ({params.file_path.absolute()})')
                async with ctx.deps.device.page.expect_file_chooser(timeout=5000) as fc_info:
                    await ctx.deps.device.page.mouse.click(x, y)
                    file_chooser = await fc_info.value
                    await file_chooser.set_files(params.file_path)
            else:
                async with ctx.deps.device.page.context.expect_page(timeout=1000) as new_page_info:
                    await ctx.deps.device.page.mouse.click(x, y)
                old_page = ctx.deps.device.page
                ctx.deps.device.page = await new_page_info.value
                await old_page.close()
        except TimeoutError:
            pass
        await JSTool.remove_highlight_position(ctx.deps.device.page)
        return ToolResult.success()

    @tool(after_delay=1)
    async def input(self, ctx: RunContext[AgentDepsType], params: InputToolParams) -> ToolResult:
        """
        在设备指定的元素中输入文本
        """
        x, y = params.get_coordinate(ctx.deps.device.device_size)
        logger.info(f'Input text: ({x}, {y}) -> {params.text}')
        await ctx.deps.device.page.mouse.click(x, y)
        await ctx.deps.device.page.keyboard.type(params.text)
        if params.send_enter:
            await ctx.deps.device.page.keyboard.press('Enter')
        return ToolResult.success()

    @staticmethod
    async def _swipe_by_mouse(
            ctx: RunContext[AgentDepsType],
            params: SwipeToolParams,
            width: int,
            height: int,
            steps: int = 1000
    ):
        if params.to == 'top':
            x1, y1, x2, y2 = 0.5 * width, 0.7 * height, 0.5 * width, 0.1 * height
        elif params.to == 'left':
            x1, y1, x2, y2 = 0.8 * width, 0.5 * height, 0.2 * width, 0.5 * height
        elif params.to == 'bottom':
            x1, y1, x2, y2 = 0.5 * width, 0.3 * height, 0.5 * width, 0.9 * height
        elif params.to == 'right':
            x1, y1, x2, y2 = 0.2 * width, 0.5 * height, 0.8 * width, 0.5 * height
        else:
            raise ValueError(f'Invalid Parameter: to={params.to}')
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        logger.info(f'Swipe from ({x1}, {y1}) to ({x2}, {y2})')
        # TODO: 禁止滑动的时候选中文字，目前先简单实现，后面寻找更优方案
        el_handle = await ctx.deps.device.page.add_style_tag(content="* {user-select: none !important;}")

        await ctx.deps.device.page.mouse.move(x1, y1)
        await ctx.deps.device.page.mouse.down()
        await ctx.deps.device.page.mouse.move(x2, y2, steps=steps)
        await ctx.deps.device.page.mouse.up()
        await JSTool.remove_element(el_handle)

    @staticmethod
    async def _swipe_by_scroll(
            ctx: RunContext[AgentDepsType],
            params: SwipeToolParams,
            width: int,
            height: int,
    ):
        if params.to == 'top':
            delta_x, delta_y = 0, 0.7 * height
        elif params.to == 'left':
            delta_x, delta_y = 0.7 * width, 0
        elif params.to == 'bottom':
            delta_x, delta_y = 0, -0.7 * height
        elif params.to == 'right':
            delta_x, delta_y = -0.7 * width, 0
        else:
            raise ValueError(f'Invalid Parameter: to={params.to}')

        logger.info(f'Scroll delta_x={delta_x}, delta_y={delta_y}')
        await ctx.deps.device.page.mouse.wheel(delta_x, delta_y)

    @tool(after_delay=1)
    async def swipe(self, ctx: RunContext[AgentDepsType], params: SwipeToolParams) -> ToolResult:
        """
        在设备屏幕中滑动或滚动
        """
        width, height = ctx.deps.device.device_size.width, ctx.deps.device.device_size.height
        has_scroll_bar = await JSTool.has_scrollbar(ctx.deps.device.page, params.to)

        if params.repeat_times is None:
            if params.expect_keywords:
                params.repeat_times = 10
            else:
                params.repeat_times = 1
        for times in range(1, params.repeat_times + 1):
            logger.info(f'swipe to {params.to}, times={times}')

            if ctx.deps.device.is_mobile and not has_scroll_bar:
                await self._swipe_by_mouse(ctx, params, width, height)
            else:
                await self._swipe_by_scroll(ctx, params, width, height)

            if params.expect_keywords:
                await asyncio.sleep(1)  # 避免滑动后截图，元素还未稳定出现
                result = await self.expect_screen_contains(ctx, params.expect_keywords)
                if result.is_success:
                    return result
                if times == params.repeat_times:
                    return ToolResult.fail(f'Expect keywords not found: {params.expect_keywords}')

        return ToolResult.success()

    @tool(after_delay=1)
    async def goback(self, ctx: RunContext[AgentDepsType], params: ToolParams) -> ToolResult:
        """
        操作返回到上一个页面
        """
        logger.info(f'go to previous page')
        await ctx.deps.device.page.go_back()
        return ToolResult.success()


class AndroidAgentTool(AgentTool):

    @staticmethod
    async def screenshot(ctx: RunContext[AgentDepsType]) -> io.BytesIO:
        image_buffer = io.BytesIO()
        screenshot = ctx.deps.device.adb_device.screenshot()
        screenshot.save(image_buffer, format='png')
        image_buffer.name = 'screen.png'
        image_buffer.seek(0)
        return image_buffer

    @tool
    async def get_screen_info(self, ctx: RunContext[AgentDepsType]) -> ToolResultWithOutput[dict]:
        """
        获取当前屏幕信息，screen_elements 包含所有解析到的元素信息，bbox 是相对值，格式为 (x1, y1, x2, y2)
        该工具禁止作为一个单独步骤
        """
        screen_info = await self.get_screen(ctx)
        return ToolResultWithOutput.success(screen_info.model_dump(include={'screen_elements'}))

    async def tear_down(self, ctx: RunContext[AgentDepsType], params: ToolParams) -> ToolResult:
        """
        任务完成或结束后的清理操作
        """
        await self.get_screen(ctx, parse_element=False)
        return ToolResult.success()

    @tool(after_delay=0)
    async def open_url(self, ctx: RunContext[AgentDepsType], params: OpenUrlToolParams) -> ToolResult:
        """
        使用设备打开URL
        """
        platform = ctx.deps.device.platform
        url_schema = get_client_url_schema(params.url, platform)
        logger.info(f'open schema: {url_schema}')

        ctx.deps.device.adb_device.shell(f'am start -a android.intent.action.VIEW -d "{url_schema}"')
        await asyncio.sleep(2)
        await self.get_screen(ctx, parse_element=False)
        return ToolResult.success()

    @tool
    async def click(self, ctx: RunContext[AgentDepsType], params: ClickToolParams) -> ToolResult:
        """
        点击设备屏幕指定的元素
        """
        x, y = params.get_coordinate(ctx.deps.device.device_size, params.position, params.offset)
        logger.info(f'Click coordinate ({x}, {y})')
        ctx.deps.device.adb_device.click(x, y)

        return ToolResult.success()

    @tool(after_delay=0)
    async def input(self, ctx: RunContext[AgentDepsType], params: InputToolParams):
        """
        在设备指定的元素中输入文本
        """
        x, y = params.get_coordinate(ctx.deps.device.device_size)
        logger.info(f'Input text: ({x}, {y}) -> {params.text}')
        ctx.deps.device.adb_device.click(x, y)
        AdbDeviceProxy(ctx.deps.device.adb_device).input_text(params.text)
        if params.send_enter:
            ctx.deps.device.adb_device.keyevent('KEYCODE_ENTER')
        return ToolResult.success()

    @tool
    async def swipe(
            self,
            ctx: RunContext[AgentDepsType],
            params: SwipeToolParams,
    ):
        """
        在设备屏幕中滑动或滚动，参数 to 表示目标方向
        """
        logger.info(f'swipe to {params.to}')
        width, height = ctx.deps.device.device_size.width, ctx.deps.device.device_size.height
        if params.to == 'top':
            x1, y1, x2, y2 = 0.5 * width, 0.7 * height, 0.5 * width, 0.1 * height
        elif params.to == 'left':
            x1, y1, x2, y2 = 0.7 * width, 0.5 * height, 0.1 * width, 0.5 * height
        elif params.to == 'bottom':
            x1, y1, x2, y2 = 0.5 * width, 0.3 * height, 0.5 * width, 0.9 * height
        elif params.to == 'right':
            x1, y1, x2, y2 = 0.3 * width, 0.5 * height, 0.9 * width, 0.5 * height
        else:
            raise ValueError(f'Invalid Parameter: to={params.to}')
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)

        if params.repeat_times is None:
            if params.expect_keywords:
                params.repeat_times = 10
            else:
                params.repeat_times = 1
        for times in range(1, params.repeat_times + 1):
            logger.info(f'Swipe from ({x1}, {y1}) to ({x2}, {y2}), times={times}')
            ctx.deps.device.adb_device.swipe(x1, y1, x2, y2, duration=2)
            if params.expect_keywords:
                await asyncio.sleep(1)  # 避免滑动后截图，元素还未稳定出现
                result = await self.expect_screen_contains(ctx, params.expect_keywords)
                if result.is_success:
                    return result
                if times == params.repeat_times:
                    return ToolResult.fail(f'Expect keywords not found: {params.expect_keywords}')

        return ToolResult.success()

    @tool
    async def swipe_from_coordinate(
            self,
            ctx: RunContext[AgentDepsType],
            params: SwipeFromCoordinateToolParams,
    ):
        """
        在设备屏幕中根据给定的坐标进行滑动操作，支持传递多个坐标进行连续滑动
        coordinates 是滑动坐标值的集合，如[(x1, y1), (x2, y2), ...]
        工具依次从坐标集中取出2组值作为开始坐标(x1, y1)和结束坐标(x2, y2)，直到完成所有坐标的滑动操作
        """
        # TODO: 先判断坐标是否在屏幕范围内
        coordinate_iter = iter(params.coordinates)
        for start_coordinate, end_coordinate in zip(coordinate_iter, coordinate_iter):
            x1, y1 = start_coordinate
            x2, y2 = end_coordinate
            logger.info(f'Swipe from ({x1}, {y1}) to ({x2}, {y2})')
            ctx.deps.device.adb_device.swipe(x1, y1, x2, y2, duration=2)
        return ToolResult.success()

    @tool
    async def open_app(
            self,
            ctx: RunContext[AgentDepsType],
            params: ToolParams,
    ):
        """
        在设备中打开或启动指定的应用(APP)
        """
        packages: list[str] = ctx.deps.device.adb_device.list_packages(filter_list=['-e'])
        sub_agent = Agent(
            ctx.model,
            output_type=str,
            system_prompt='你是一个移动端应用助手，负责根据用户输入的指令从提供的应用包名列表找出用户指令对应的包名，并仅返回包名，如果都不匹配则返回空字符串'
        )
        prompt = (f'用户指令：{params.instruction}\n'
                  f'应用包名列表：{packages}')
        result = await sub_agent.run(prompt, output_type=str)
        package_name = result.output
        if not package_name:
            return ToolResultWithOutput.failed(output='在该设备中未找到对应的应用')
        logger.info(f'Find App package name：{package_name}')
        ctx.deps.device.adb_device.app_start(package_name)
        await asyncio.sleep(2)
        await self.get_screen(ctx, parse_element=False)
        return ToolResult.success()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 18:30
from collections import OrderedDict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, TypeVar, Literal, Generic, TypeAlias

from pydantic import BaseModel, Field, confloat, conlist, ConfigDict

from .config import Settings

T = TypeVar('T')
DeviceT = TypeVar('DeviceT')
ToolT = TypeVar('ToolT')

SimulateDeviceType: TypeAlias = Literal['iPhone 15', 'iPhone 15 Pro', 'iPhone 15 Pro Max', 'iPhone 6'] | str


class DeviceSize(BaseModel):
    """当前设备尺寸"""
    width: int
    height: int


class ScreenInfo(BaseModel):
    """当前屏幕信息"""
    image_url: str = ''
    screen_elements: list[dict] = Field(default_factory=list)

    def reset(self):
        self.image_url = ''
        self.screen_elements = []


class StepInfo(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    step: int = 0
    description: str = ''
    action: str = ''
    params: dict = Field(default_factory=dict)
    image_url: str = ''
    planning: Optional['PlanningStep'] = Field(default=None, exclude=True)
    screen_elements: list[dict] = Field(default_factory=list)
    parallel_tool_calls: bool = Field(default=False, exclude=True)
    is_success: bool = True


@dataclass
class AgentContext:
    """
    Agent 额外的上下文信息
    steps: 所有步骤信息
    current_step: 当前步骤信息
    screen_info: 当前屏幕信息
    """
    steps: OrderedDict[int, StepInfo] = field(default_factory=OrderedDict)
    current_step: StepInfo = field(default_factory=StepInfo)

    def set_step_failed(self, reason: str):
        self.steps[self.current_step.step].is_success = False
        self.steps[self.current_step.step].action = 'mark_failed'
        self.steps[self.current_step.step].params = {'reason': reason}

    def add_step_info(self, step_info: StepInfo):
        self.current_step = self.steps.setdefault(step_info.step, step_info)
        return self.current_step

    def update_step_info(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self.current_step, key):
                setattr(self.current_step, key, value)
        return self.current_step


@dataclass
class AgentDeps(Generic[DeviceT, ToolT]):
    settings: Settings
    device: DeviceT
    tool: ToolT
    context: AgentContext = field(default_factory=AgentContext)


class ToolParams(BaseModel):
    instruction: str = Field(description='用户指令，描述该步骤要做什么', exclude=True)
    action: str = Field(description='要执行的动作，如 click、input、swipe 等, 是调用的工具名称')


class OpenUrlToolParams(ToolParams):
    url: str = Field(description='要打开的url网址')


PositionType: TypeAlias = Literal['left', 'right', 'top', 'bottom']


class LocationToolParams(ToolParams):
    element_bbox: conlist(confloat(ge=0.0, le=1.0), min_length=4, max_length=4) = Field(description='要操作的元素 bbox')
    element_content: str = Field(description='要操作的元素内容')

    def get_coordinate(self,
                       device_size: DeviceSize,
                       position: Optional[PositionType] = None,
                       offset: Optional[float] = None) -> tuple[int, int]:
        """计算坐标, 返回 (x, y)"""
        x1, y1, x2, y2 = self.element_bbox
        width, height = device_size.width, device_size.height
        x, y = (x1 + x2) / 2, (y1 + y2) / 2
        offset = 0.25 if offset is None else offset
        if position == 'left':
            x = x1 + (x2 - x1) * offset
        elif position == 'right':
            x = x2 - (x2 - x1) * offset
        elif position == 'top':
            y = y1 + (y2 - y1) * offset
        elif position == 'bottom':
            y = y2 - (y2 - y1) * offset

        return int(x * width), int(y * height)


class ClickToolParams(LocationToolParams):
    """示例：
    - 点击"确定"按钮 -> position=None, offset=None
    - 点击"确定"按钮左侧 -> position='left', offset=None
    - 点击"确定"按钮左侧1/2处 -> position='left', offset=0.5
    - 点击"上传"按钮，上传文件:"/Users/Desktop/1.png" -> position=None, offset=None, file_path='/Users/Desktop/1.png'
    """
    position: Optional[PositionType] = Field(default=None, description='点击元素的相对位置')
    offset: Optional[float] = Field(default=None, description='相对位置的偏移量')
    file_path: Optional[Path] = Field(default=None, description='要上传的文件路径')


class InputToolParams(LocationToolParams):
    """
    示例：
    - 输入"123456" -> text="123456"，send_enter=True
    - 输入"123456"，不发送回车键 -> text="123456"，send_enter=False
    """
    text: str = Field(description='要输入的文本')
    send_enter: bool = Field(default=True, description='是否发送回车键')


class SwipeToolParams(ToolParams):
    """
    示例：
    向上滑动 2 次 -> to='top', repeat_times=2
    向上滑动最多 5 次，直到页面中出现 "确定" 元素 -> to='top', repeat_times=5, expect_keywords=['确定']
    向上滑动，直到出现"胡广生"元素 -> to='top', repeat_times=None, expect_keywords=['胡广生']
    """
    to: Literal['left', 'right', 'top', 'bottom'] = Field(description='滑动方向')
    repeat_times: Optional[int] = Field(default=1, description='重复次数或最多重复次数，默认为1次')
    expect_keywords: Optional[list[str]] = Field(description='期望出现的关键字列表')


class SwipeFromCoordinateToolParams(ToolParams):
    coordinates: conlist(item_type=tuple[int, int], min_length=2) = Field(description='滑动起始坐标列表')


class WaitToolParams(ToolParams):
    """
    示例：
    等待2秒 -> timeout=2，expect_keywords=None
    等待5秒，直到出现"确定"按钮 -> timeout=5，expect_keywords=['确定']
    """
    timeout: int = Field(description='等待时间，单位为秒')
    expect_keywords: Optional[list[str]] = Field(description='期望出现的关键字列表')


class AssertContainsParams(ToolParams):
    expect_keywords: list[str] = Field(description='期望包含的关键字列表')


class AssertNotContainsParams(ToolParams):
    unexpect_keywords: list[str] = Field(description='期望不包含的关键字列表')


class MarkFailedParams(BaseModel):
    reason: str = Field(description='失败原因')


class ToolResult(BaseModel, Generic[T]):
    is_success: bool

    @classmethod
    def success(cls):
        return cls(is_success=True)

    @classmethod
    def failed(cls):
        return cls(is_success=False)


class ToolResultWithOutput(ToolResult, Generic[T]):
    output: Optional[T] = Field(default=None)

    @classmethod
    def success(cls, output: T = None):
        return cls(is_success=True, output=output)

    @classmethod
    def failed(cls, output: T = None):
        return cls(is_success=False, output=output)


class PlanningStep(BaseModel):
    instruction: str = Field(description='步骤指令，用一句话描述该步骤要做什么')


class PlanningOutputType(BaseModel):
    """
    用户指令分解后，规划出的指令序列
    """
    steps: list[PlanningStep] = Field(description='步骤序列')


class StepOutputType(BaseModel):
    """
    用户指令完成后判断是否成功
    """
    is_success: bool = Field(description='所有操作是否成功')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/6/19 11:46
import asyncio
import hashlib
import imghdr
import io
import os
import time
from abc import ABC, abstractmethod
from io import BytesIO, StringIO
from typing import IO, Union, Any

from PIL import Image
from loguru import logger
from minio import Minio, S3Error
from qcloud_cos.cos_client import CosS3Client, CosConfig


class TinyImg:
    def __init__(self, fp: Union[IO, BytesIO, StringIO, Any]):
        self.fp = fp

    def is_image(self):
        location = self.fp.tell()
        h = self.fp.read(32)
        self.fp.seek(location)
        for tf in imghdr.tests:
            res = tf(h, self.fp)
            if res:
                return True
        return False

    def to_webp(self):
        """WebP 图像格式的最大分辨率为16383 x 16383"""
        location = self.fp.tell()
        try:
            img = Image.open(self.fp)
            max_size = max(img.size)
            if max_size > 16383:
                ratio = 16383 / max_size
                resize = (int(img.size[0] * ratio), int(img.size[1] * ratio))
                logger.warning(f'图片宽高超过16383，将压缩图片大小：{img.size} -> {resize}')
                img = img.resize(resize)
            file = io.BytesIO()
            img.save(file, 'webp')
            self.fp.seek(location)
            file.seek(0)
            return file
        except Exception as e:
            error_file = f'{time.time()}_error_file.png'
            self.fp.seek(location)
            with open(error_file, 'wb') as f:
                # noinspection PyTypeChecker
                f.write(self.fp.read())
            logger.error(f'图片转换失败：错误文件 {error_file}')
            raise e


# 策略接口
class StorageStrategy(ABC):
    @abstractmethod
    def upload_file(self, file, prefix='', suffix='.png'):
        pass

    @abstractmethod
    async def async_upload_file(self, file, prefix='', suffix='.png'):
        pass

    @staticmethod
    def get_file_md5(file):
        file.seek(0)
        m = hashlib.md5()
        m.update(file.read())
        file.seek(0)
        return m.hexdigest()


# COS策略实现
class CosStrategy(StorageStrategy):
    def __init__(self, secret_id, secret_key, region, endpoint, bucket):
        _cos_config = CosConfig(Region=region, SecretId=secret_id, SecretKey=secret_key, Endpoint=endpoint)
        self._client = CosS3Client(_cos_config)
        self.bucket = bucket

    def upload_file(self, file, prefix='', suffix='.png'):
        file_md5 = self.get_file_md5(file)
        key = f'{prefix}{file_md5}{suffix}'

        try:
            if not self._client.object_exists(self.bucket, key):
                file = TinyImg(file).to_webp() if suffix == '.png' else file
                self._client.put_object(Bucket=self.bucket, Key=key, Body=file)
            cos_url = self._client.get_object_url(self.bucket, key)
            return cos_url
        except Exception as e:
            logger.error(f'上传文件失败：{e}')
            raise e

    async def async_upload_file(self, file, prefix='', suffix='.png'):
        return await asyncio.to_thread(self.upload_file, file, prefix=prefix, suffix=suffix)


# MinIO策略实现
class MinioStrategy(StorageStrategy):
    def __init__(self, access_key, secret_key, endpoint, bucket, region=None, secure=False):
        self._client = Minio(access_key=access_key, secret_key=secret_key, region=region, endpoint=endpoint,
                             secure=secure)
        self.endpoint = endpoint
        self.bucket = bucket
        self.protocol = 'https' if secure else 'http'

    def object_exists(self, key):
        try:
            self._client.stat_object(bucket_name=self.bucket, object_name=key)
            return True
        except S3Error as err:
            if err.code == "NoSuchKey":
                return False
            logger.error(f"MinIO服务异常: {err}")
        return False

    def upload_file(self, file, prefix='', suffix='.png'):
        file_md5 = self.get_file_md5(file)
        key = f'{prefix}{file_md5}{suffix}'

        try:
            if not self.object_exists(key):
                file = TinyImg(file).to_webp() if suffix == '.png' else file
                # 获取文件大小
                file.seek(0, os.SEEK_END)
                file_size = file.tell()
                file.seek(0)
                self._client.put_object(bucket_name=self.bucket, object_name=key, data=file, length=file_size)

            return f"{self.protocol}://{self.endpoint}/{self.bucket}/{key}"
        except Exception as e:
            logger.error(f"上传文件失败: {e}")
            raise e

    async def async_upload_file(self, file, prefix='', suffix='.png'):
        return await asyncio.to_thread(self.upload_file, file, prefix=prefix, suffix=suffix)


# 主要的存储客户端类
class StorageClient:
    def __init__(self, strategy: StorageStrategy):
        self._strategy = strategy

    def __repr__(self):
        return f"StorageClient(strategy={self._strategy})"

    @classmethod
    def create_from_config(cls, cos_config, minio_config):
        """根据配置创建存储客户端"""
        # 优先使用COS，如果COS配置完整
        if cos_config.secret_id and cos_config.secret_key:
            strategy = CosStrategy(
                secret_id=cos_config.secret_id,
                secret_key=cos_config.secret_key,
                region=cos_config.region,
                endpoint=cos_config.endpoint,
                bucket=cos_config.bucket
            )
        # 如果COS配置不完整，使用MinIO
        elif minio_config.access_key and minio_config.secret_key:
            strategy = MinioStrategy(
                access_key=minio_config.access_key,
                secret_key=minio_config.secret_key,
                endpoint=minio_config.endpoint,
                bucket=minio_config.bucket,
                region=minio_config.region,
                secure=minio_config.secure
            )
        else:
            raise ValueError("未找到有效的存储配置，请检查COS或MinIO环境变量配置")

        return cls(strategy)

    def upload_file(self, file, prefix='', suffix='.png'):
        return self._strategy.upload_file(file, prefix, suffix)

    async def async_upload_file(self, file, prefix='', suffix='.png'):
        return await self._strategy.async_upload_file(file, prefix, suffix)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 15:31
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional, Union
from random import randint

from loguru import logger
from pydantic import TypeAdapter
from pydantic_ai import Agent, UserPromptNode, ModelRequestNode, CallToolsNode, RunContext, UnexpectedModelBehavior
from pydantic_ai.agent import AgentRunResult
from pydantic_ai.messages import ToolReturnPart, ToolCallPart
from pydantic_ai.usage import Usage

from .config import global_settings, model_settings
from .deps import AgentDeps, SimulateDeviceType, PlanningOutputType, StepOutputType, PlanningStep, ToolParams, StepInfo, \
    MarkFailedParams
from .device import AndroidDevice, WebDevice
from .prompt import SYSTEM_PROMPT, PLANNING_SYSTEM_PROMPT
from .tools import AndroidAgentTool, WebAgentTool, AgentDepsType
from .util.platform import Platform


@dataclass
class PlanningAgent:
    """PlanningAgent class for planning tasks."""
    model: Optional[str] = None
    deps: Optional[AgentDepsType] = None

    async def run(self, prompt: str) -> AgentRunResult[PlanningOutputType]:
        """Run the agent with the given prompt."""
        model = self.model or global_settings.model
        agent = Agent(
            model=model,
            system_prompt=PLANNING_SYSTEM_PROMPT,
            output_type=PlanningOutputType,
            model_settings=model_settings
        )
        return await agent.run(prompt.strip(), deps=self.deps)


@dataclass
class UiAgent:
    model: str
    deps: AgentDepsType
    agent: Agent[AgentDepsType]

    @classmethod
    async def create(cls, *args, **kwargs):
        """Async factory method to create an instance of this class."""
        raise NotImplementedError

    @staticmethod
    async def create_report(report_data: str, report_dir: Union[Path, str]) -> Path:
        """Create a report file based on the given data and directory."""
        logger.info('创建步骤报告...')
        logger.debug(f'report_data: {report_data}')

        report_dir = Path(report_dir)
        report_dir.mkdir(parents=True, exist_ok=True)

        template = Path(__file__).parent / "report_template.html"
        content = template.read_text(encoding='utf-8').replace('{reportData}', report_data)
        output_path = report_dir / f'report_{datetime.now():%Y%m%d%H%M%S}_{randint(10000, 99999)}.html'
        output_path.write_text(content, encoding='utf-8')
        logger.info(f"报告：{output_path.resolve().as_uri()}")
        return output_path

    def handle_graph_node(self, node):
        """Format the logger node based on the given node type."""
        if isinstance(node, UserPromptNode):
            logger.log('DETAIL', f"🤖Agent start user task: {repr(node.user_prompt)}")

        elif isinstance(node, ModelRequestNode):
            for part in node.request.parts:
                if isinstance(part, ToolReturnPart):
                    logger.log('DETAIL', f"🤖Agent tool result: {part.tool_name} -> {part.content}")

        elif isinstance(node, CallToolsNode):
            parts = node.model_response.parts
            tool_parts = [part for part in parts if isinstance(part, ToolCallPart)]
            self.deps.context.current_step.parallel_tool_calls = False
            self.deps.context.current_step.parallel_tool_calls = len(tool_parts) > 1
            for part in tool_parts:
                logger.log('DETAIL', f"🤖Agent tool call: {part.tool_name}, args: {part.args}")

    async def _sub_agent_run(self, planning, usage) -> AgentRunResult:
        async with self.agent.iter(user_prompt=planning.instruction, deps=self.deps, usage=usage) as agent_run:
            async for node in agent_run:
                self.handle_graph_node(node)
            return agent_run.result

    async def run(self, prompt: str, system_prompt: Optional[str] = None, report_dir: str = "./report"):
        # TODO: 给用户添加额外的自定义系统提示词，某些场景需要，如：如果出现位置、权限、用户协议等弹窗，点击同意。如果出现登录页面，关闭它。
        logger.info(f"🤖Agent start planning...")

        planning_agent = PlanningAgent(model=self.model, deps=self.deps)
        planning_result = await planning_agent.run(prompt)
        planning_steps = planning_result.output.steps

        planning_steps = [*planning_steps, PlanningStep(instruction='结束任务')]
        logger.info(f"🤖Agent planning finished.")
        for index, step in enumerate(planning_steps, 1):
            logger.info(f'◽️step{index}. {step.instruction}')

        if system_prompt:
            self.agent.system_prompt(lambda: system_prompt)

        usage = planning_result.usage()
        ctx = RunContext(deps=self.deps, model=self.agent.model, usage=Usage(), prompt=None)

        logger.info(f"🤖Agent start executing steps...")
        for step, planning in enumerate(planning_steps, start=1):
            self.deps.context.add_step_info(StepInfo(step=step, planning=planning, description=planning.instruction))
            logger.info('')
            logger.info(f'▶️ step={step} {planning.instruction}')

            if planning.instruction != '结束任务':
                try:
                    result = await self._sub_agent_run(planning, usage)
                    usage = result.usage()
                    logger.info(f"💬 {result.output}")
                except UnexpectedModelBehavior as e:
                    await self.deps.tool.mark_failed(ctx, MarkFailedParams(
                        reason=str(e),
                    ))
                    logger.error(f'step={step} {planning.instruction}: {e}')

                logger.info(f'{"✅" if self.deps.context.current_step.is_success else "❌"} '
                            f'step={step} {planning.instruction}')
            else:
                await self.deps.tool.tear_down(ctx, params=ToolParams(action='tear_down', instruction='任务完成'))

            # 步骤执行后如果没有截图则自动补上，比如滑动、等待
            if not self.deps.context.current_step.image_url:
                await self.deps.tool.get_screen(ctx, parse_element=False)

            if not self.deps.context.current_step.is_success:
                break

        logger.debug(f"steps: {self.deps.context.steps}")
        logger.log('DETAIL', f"usage: {usage}")

        is_success_output = all([step.is_success for step in self.deps.context.steps.values()])

        report_data = {'is_success': is_success_output,
                       'device_size': self.deps.device.device_size,
                       'steps': self.deps.context.steps}
        report_json = TypeAdapter(dict).dump_json(report_data).decode(encoding='utf-8')
        report_path = await self.create_report(report_json, report_dir)

        steps_output = [
            step.model_dump(include={'step', 'description', 'action', 'is_success'})
            for step in self.deps.context.steps.values()
        ]

        return {
            'is_success': is_success_output,
            'steps': steps_output,
            'report_path': report_path
        }


class WebAgent(UiAgent):
    """WebAgent class for web automation."""

    @classmethod
    async def create(
            cls,
            model: Optional[str] = None,
            *,
            device: Optional[WebDevice] = None,
            simulate_device: Optional[SimulateDeviceType] = None,
            headless: Optional[bool] = None,
            tool_cls: Optional[type[WebAgentTool]] = None,
            debug: Optional[bool] = None,
    ):
        settings = global_settings.copy_and_update(
            model=model,
            simulate_device=simulate_device,
            headless=headless,
            debug=debug)

        logger.info(f'settings: {settings}')

        device = device or await WebDevice.create(settings.headless, settings.simulate_device)
        tool = WebAgentTool() if tool_cls is None else tool_cls()
        deps: AgentDeps[WebDevice, WebAgentTool] = AgentDeps(settings, device, tool)

        agent = Agent[AgentDeps](
            model=settings.model,
            system_prompt=SYSTEM_PROMPT,
            model_settings=model_settings,
            deps_type=AgentDeps,
            tools=tool.tools,
            retries=3
        )
        return cls(model, deps, agent)


class MobileAgent(UiAgent):
    """MobileAgent class for mobile device automation."""

    @classmethod
    async def create(
            cls, model: Optional[str] = None,
            *,
            serial: Optional[str] = None,
            platform: Optional[str | Platform] = None,
            tool_cls: Optional[type[AndroidAgentTool]] = None,
            debug: Optional[bool] = None,
    ):
        settings = global_settings.copy_and_update(model=model, debug=debug)

        logger.info(f'settings: {settings}')

        device = await AndroidDevice.create(serial=serial, platform=platform)

        tool = AndroidAgentTool() if tool_cls is None else tool_cls()
        deps: AgentDeps[AndroidDevice, AndroidAgentTool] = AgentDeps(settings, device, tool)

        agent = Agent[AgentDeps](
            model=settings.model,
            system_prompt=SYSTEM_PROMPT,
            model_settings=model_settings,
            deps_type=AgentDeps,
            tools=tool.tools,
            retries=2
        )
        return cls(model, deps, agent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 16:52

PLANNING_SYSTEM_PROMPT = """
## 角色定位
「高精度UI操作规划专家」：专注准确解析用户意图，将用户指令转化为可执行的原子化操作步骤

## 目标
指令分解：将复杂指令拆解为多个原子化操作步骤，每个步骤必须满足：  
- 直接关联用户意图  
- 单个步骤只能进行一个操作

## 约束
❗强制要求：  
- 所有操作必须按照指令的顺序进行规划 

## 示例
- 点击"close"关闭弹窗，若弹窗不存在则跳过 -> {'instruction': '点击"close"关闭弹窗，若"close"元素不存在则跳过'}
- 向上滑动3次 -> {'instruction': '向上滑动 3 次'}
- 打开QQ音乐APP -> {'instruction': '打开 "QQ音乐" APP'}
- 点击"上传"按钮，上传文件"/Users/pic.png" -> {'instruction': '点击"上传"按钮，上传文件: "/Users/pic.png"'}
"""

SYSTEM_PROMPT = """
## 角色定位
「高精度UI操作执行专家」：专注准确解析用户意图，严格遵循屏幕实时状态执行可靠操作

## 目标
1. 指令分解：直接关联用户意图，将指令拆解为原子化操作序列
2. 根据拆解的指令使用相应的工具进行操作

## 工作流程
1. 如果需要获取屏幕元素信息，则调用 `get_screen_info` 工具获取当前屏幕元素信息
2. 根据用户指令和当前屏幕的元素信息查找目标元素，如果未找到，则调用 `mark_failed` 工具标记失败
3. 调用相应的工具执行操作
4. 如果用户指令未完成，则重复以上1-3操作步骤

## 元素查找
- 优先根据元素文本内容、图标、位置等特征匹配目标元素
- 若文本内容模糊或不完整，可结合上下文、元素类型辅助定位
- 元素信息上下文中包含每个元素上、下、左、右相邻的元素id信息，可结合用户指令找到相邻的目标元素，比如 left_elem_ids: [1, 2, 3]，"点击xx元素左测的第一个元素"，则目标元素是id为1的元素
    - left_elem_ids -> 左侧相邻的元素id列表
    - right_elem_ids -> 右侧相邻的元素id列表
    - top_elem_ids -> 上方相邻的元素id列表
    - bottom_elem_ids -> 下方相邻的元素id列表

## 补充说明
- "打开url"、"打开app"、"打开应用"、"等待"、"滑动"、"滚动"操作不需要获取界面元素信息，其它操作必须先获取界面元素信息  
- 过程中如果出现位置、权限、用户协议、广告等弹窗等，则调用 `click` 工具点击"同意"、"确定"、"允许"、"跳过"、"关闭"、"取消"、"我知道了"、"Dismiss"、"X"等按钮

## 工具所需要的关键参数描述
- action：要执行的动作，如 click、input 等，是调用的工具名称
- element_bbox：要操作的元素 bbox，格式为 [x1, y1, x2, y2]，x1, y1, x2, y2 是相对屏幕分辨率的坐标值，取值范围为 [0-1] 

## 约束
❗强制要求：
- 所有操作必须按照指令的意图进行
- 如果屏幕中找到多个相同元素，且用户没有明确要求，则优先选择第一个
- 每次只能同时调用一个工具，如果需要多个工具，则需要先完成当前工具的执行，再调用下一个工具
- 在当前屏幕中元素未找到或无法进行操作时，必须先调用 `mark_failed` 工具标记失败，然后结束任务

❗绝对禁止：
- 假设屏幕外或历史状态的元素
- 添加用户指令未明确要求的操作
"""

SYSTEM_PROMPT_EN = """
## Role Positioning
「High-Precision UI Operation Expert」: Focuses on accurately interpreting user intent and strictly executes reliable
operations based on real-time screen status

## Objectives
1. Instruction Decomposition: Break down complex instructions into atomic operation sequences, each step must satisfy:
    - Directly correlate with user intent
    - Have explicitly verifiable conditions
    - Possess corresponding screen action support
2. Use appropriate tools to execute operations based on decomposed instructions

## Workflow
1. Obtain action information from current device screen
2. Locate target elements based on user instructions and current screen action information
3. Invoke corresponding tools to perform operations
4. Repeat steps 1-3 if user instructions remain uncompleted

## Device Information
- Device Name: {device_name}
- Screen Resolution: {screen_resolution}

## Constraints
❗Mandatory Requirements:
- All planned operations must be based on device screen action information, with coordinates calculated according to
current screen - operation fails if relevant elements are not found
- When multiple identical elements exist on screen, prioritize selecting the first one
- All operations must follow instruction sequence
- Only invoke one tool per operation
- Mandatory retry after operation failure - return error message and terminate task if retry fails

❗Absolute Prohibitions:
- Assume elements outside screen or in historical states
- Add operations not explicitly required by user instructions

## Results
Output execution results to "results" after completing instructions:
- Step sequence number → step
- Step description → description
- Executed action → action
- Element ID → element_id
- Element information → element_bbox
- Error message per step → error (empty if step succeeds)
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/6/6 14:58
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from adbutils import AdbClient, AdbDevice
from playwright.async_api import async_playwright, Playwright, Page, BrowserContext, ViewportSize

from .deps import DeviceSize
from .util.platform import Platform


@dataclass
class WebDevice:
    playwright: Optional[Playwright]
    context: BrowserContext
    page: Page
    device_size: DeviceSize
    simulate_device: Optional[str] = None
    is_mobile: Optional[bool] = None

    @classmethod
    async def create(cls, headless: bool = False, simulate_device: Optional[str] = None) -> "WebDevice":
        """异步工厂方法用于创建实例"""
        playwright = await async_playwright().start()
        context_params = {'viewport': ViewportSize(width=1600, height=900)}
        is_mobile = False
        if simulate_device and simulate_device in playwright.devices:
            is_mobile = True
            context_params.update(playwright.devices[simulate_device])
            del context_params['has_touch']  # fix swipe scene
            del context_params['default_browser_type']  # launch_persistent_context not support default_browser_type

        context = await playwright.chromium.launch_persistent_context(
            user_data_dir=Path(tempfile.gettempdir()) / 'playwright',
            channel='chrome',
            headless=headless,
            # devtools=True,
            **context_params
        )

        page = context.pages[0]
        device_size = DeviceSize(**page.viewport_size)

        return cls(playwright, context, page, device_size, simulate_device, is_mobile)

    @classmethod
    async def from_page(cls, page: Page, simulate_device: Optional[str] = None) -> "WebDevice":
        """通过page对象创建实例"""
        playwright = None
        context = page.context

        device_size = DeviceSize(**page.viewport_size)

        return cls(playwright, context, page, device_size, simulate_device)


@dataclass
class AndroidDevice:
    client: AdbClient
    adb_device: AdbDevice
    device_size: DeviceSize
    platform: Platform

    @classmethod
    async def create(cls, serial: Optional[str] = None, platform: Optional[Platform] = Platform.QY):
        """异步工厂方法用于创建实例"""
        client = AdbClient()
        current_devices = client.device_list()
        if serial:
            if serial not in current_devices:
                output = client.connect(serial, timeout=10)
                if 'connected' not in output:
                    raise Exception(f"adb connect failed: {output}")
            adb_device: AdbDevice = client.device(serial=serial)
        elif current_devices:
            adb_device: AdbDevice = client.device_list()[0]
        else:
            raise Exception("No adb device found")

        window_size = adb_device.window_size()
        device_size = DeviceSize(width=window_size.width, height=window_size.height)
        return cls(client, adb_device, device_size, platform)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/11/19 12:05
from openai import OpenAI

client = OpenAI(
    api_key="EMPTY",
    base_url="http://21.91.23.205:8000/v1",
    timeout=3600
)

# Task-specific base prompts
TASKS = {
    "ocr": "OCR:",
    "table": "Table Recognition:",
    "formula": "Formula Recognition:",
    "chart": "Chart Recognition:",
}

messages = [
    {
        "role": "user",
        "content": [
            {
                "type": "image_url",
                "image_url": {
                    "url": "https://tme-dev-test-cos-1257943044.cos-internal.ap-guangzhou.tencentcos.cn/page-shot/image/da2ce0563235c1337a22abb54ea56f4d.png"
                }
            },
            {
                "type": "text",
                "text": TASKS["table"]
            }
        ]
    }
]

response = client.chat.completions.create(
    model="PaddlePaddle/PaddleOCR-VL",
    messages=messages,
    temperature=0.0,
)
print(f"{response.choices[0].message.content}")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/23 16:52

PLANNING_SYSTEM_PROMPT = """
## 角色定位
「高精度UI操作规划专家」：专注准确解析用户意图，将用户指令转化为可执行的原子化操作步骤

## 目标
指令分解：将复杂指令拆解为多个原子化操作步骤，每个步骤必须满足：  
- 直接关联用户意图  
- 单个步骤进行一个操作
- 保留原始指令中的所有信息

## 约束
❗强制要求：  
- 所有操作必须按照指令的顺序进行规划 

## 示例
- 点击"close"关闭弹窗，若弹窗不存在则跳过 -> {'instruction': '点击"close"关闭弹窗，若"close"元素不存在则跳过'}
- 向上滑动3次 -> {'instruction': '向上滑动 3 次'}
- 向上滑动，直到出现"登录"按钮 -> {'instruction': '向上滑动，直到出现"登录"按钮'}
- 打开QQ音乐APP -> {'instruction': '打开 "QQ音乐" APP'}
- 点击"上传"按钮，上传文件"/Users/pic.png" -> {'instruction': '点击"上传"按钮，上传文件: "/Users/pic.png"'}
"""

SYSTEM_PROMPT = """
## 角色定位
「高精度UI操作执行专家」：基于实时屏幕状态，精准解析用户意图并执行可靠的原子化操作

## 核心目标
1. **意图理解**：准确解析用户指令，将复杂任务拆解为可执行的原子化操作序列
2. **精准执行**：严格基于当前屏幕实际状态，调用相应工具完成操作
3. **异常处理**：识别并妥善处理执行过程中的异常情况

## 工作流程
1. 如果需要获取屏幕元素信息，则调用 `get_screen_info` 工具获取当前屏幕元素信息
2. 根据用户指令和当前屏幕的元素信息查找目标元素，如果未找到，则调用 `mark_failed` 工具标记失败
3. 调用相应的工具执行操作
4. 如果用户指令未完成，则重复以上1-3操作步骤
5. 工具返回结果包含 `is_success` 字段，True表示当前操作成功，False表示当前操作失败

## 补充说明
- "打开url"、"打开app"、"打开应用"、"等待"、"滑动"、"滚动"操作不需要获取界面元素信息，其它操作必须先获取界面元素信息  
- 过程中如果出现位置、权限、用户协议、广告等弹窗等，则调用 `click` 工具点击"同意"、"确定"、"允许"、"跳过"、"关闭"、"取消"、"我知道了"、"Dismiss"、"X"等按钮
- `input` 工具内部会自动调用 `click` 工具进行输入框激活操作，因此可跳过输入前的激活操作

## 元素定位规范
### 定位策略
1. 文本内容精确匹配（优先级最高）
2. 如果文本内容不能精确匹配，则根据用户的指令进行模糊匹配（见"模糊匹配策略"）
3. 相邻元素关系推断（见"空间关系定位"）

**失败处理**：若目标元素未找到，立即调用 `mark_failed` 工具标记失败并终止任务

### 模糊匹配策略
- 当文本内容不完整时，结合元素类型（button、input等）和上下文语义进行匹配
- 优先选择语义相关性最高的元素
- 若存在多个候选元素，选择屏幕中首次出现的元素（从上到下、从左到右）

### 空间关系定位
元素上下文包含四个方向的相邻元素ID列表，可用于相对位置定位：
- `left_elem_ids`：左侧相邻元素ID列表（从近到远排序）
- `right_elem_ids`：右侧相邻元素ID列表（从近到远排序）
- `top_elem_ids`：上方相邻元素ID列表（从近到远排序）
- `bottom_elem_ids`：下方相邻元素ID列表（从近到远排序）

**示例**：
1. 指令"点击搜索框右侧的第1个按钮"  
- 若搜索框的 `right_elem_ids: [5, 8, 12]`
- 则目标元素为right_elem_ids对应索引0，即 id=5 的元素

2. 指令"点击搜索框右侧的第2个按钮"  
- 若搜索框的 `right_elem_ids: [5, 8, 12]`
- 则目标元素为right_elem_ids索引1，即 id=8 的元素

3. 指令"点击搜索框左侧的按钮"  
- 若搜索框的 `left_elem_ids: [3, 2, 1]`
- 因无明确左侧第几个元素，则默认取目标元素left_elem_ids索引0， 即 id=3 的元素

## 弹窗自动处理
在执行主任务过程中，若遇到以下元素，应自动处理：
- **权限请求**：点击"允许"、"同意"、"确定"
- **用户协议**：点击"同意"、"接受"
- **广告弹窗**：点击"跳过"、"关闭"、"X"、"Dismiss"
- **通知提示**：点击"我知道了"、"确定"、"取消"

## 工具参数规范
### 通用参数
- `action`：操作类型标识符，对应工具名称（如：click、input、scroll）
- `element_bbox`：元素边界框坐标，格式为 `[x1, y1, x2, y2]`
  - 坐标系统：归一化坐标（相对屏幕分辨率）
  - 取值范围：[0.0, 1.0]
  - 示例：`[0.1, 0.2, 0.3, 0.4]` 表示元素左上角在屏幕 (10%, 20%) 位置，右下角在 (30%, 40%) 位置

## 执行约束
### 强制要求 ❗
1. **意图忠实性**：所有操作必须严格遵循用户指令意图，不得擅自添加或修改
2. **状态依赖性**：所有操作必须基于当前屏幕实际状态，禁止假设屏幕外或历史状态的元素
3. **顺序执行性**：严格按照指令顺序执行，每次仅调用一个工具
4. **元素唯一性**：当屏幕存在多个相同元素且用户未明确指定时，选择首个匹配元素
5. **失败即停原则**：元素未找到或操作无法执行时，必须调用 `mark_failed` 标记失败并立即终止

### 绝对禁止 ❌
1. 假设或推测屏幕外、历史状态或未来状态的元素
2. 添加用户指令中未明确要求的操作步骤
3. 同时调用多个工具（必须等待当前工具执行完成）
4. 忽略元素定位失败，强行执行后续操作
5. 修改用户指令的原始意图或执行顺序
"""

SYSTEM_PROMPT_EN = """
## Role Positioning
「High-Precision UI Operation Expert」: Focuses on accurately interpreting user intent and strictly executes reliable
operations based on real-time screen status

## Objectives
1. Instruction Decomposition: Break down complex instructions into atomic operation sequences, each step must satisfy:
    - Directly correlate with user intent
    - Have explicitly verifiable conditions
    - Possess corresponding screen action support
2. Use appropriate tools to execute operations based on decomposed instructions

## Workflow
1. Obtain action information from current device screen
2. Locate target elements based on user instructions and current screen action information
3. Invoke corresponding tools to perform operations
4. Repeat steps 1-3 if user instructions remain uncompleted

## Device Information
- Device Name: {device_name}
- Screen Resolution: {screen_resolution}

## Constraints
❗Mandatory Requirements:
- All planned operations must be based on device screen action information, with coordinates calculated according to
current screen - operation fails if relevant elements are not found
- When multiple identical elements exist on screen, prioritize selecting the first one
- All operations must follow instruction sequence
- Only invoke one tool per operation
- Mandatory retry after operation failure - return error message and terminate task if retry fails

❗Absolute Prohibitions:
- Assume elements outside screen or in historical states
- Add operations not explicitly required by user instructions

## Results
Output execution results to "results" after completing instructions:
- Step sequence number → step
- Step description → description
- Executed action → action
- Element ID → element_id
- Element information → element_bbox
- Error message per step → error (empty if step succeeds)
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/6/24 12:18
from typing import Literal, Optional

from loguru import logger
from pydantic import Field
from pydantic_ai.settings import ModelSettings
from pydantic_settings import BaseSettings, SettingsConfigDict

from .util.storage import StorageClient


class CosConfig(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', env_prefix='cos_', extra='ignore')

    region: str = Field(default='')
    secret_id: str = Field(default='')
    secret_key: str = Field(default='')
    endpoint: str = Field(default='')
    bucket: str = Field(default='')


class MinioConfig(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', env_prefix='minio_', extra='ignore')

    access_key: str = Field(default='')
    secret_key: str = Field(default='')
    endpoint: str = Field(default='')
    bucket: str = Field(default='')
    region: str = Field(default='')
    secure: bool = Field(default=False)


def create_storage_client():
    """创建存储客户端的工厂函数"""
    cos_config = CosConfig()
    minio_config = MinioConfig()
    return StorageClient.create_from_config(cos_config, minio_config)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', env_prefix='agent_', extra='ignore')

    headless: Optional[bool] = True
    model: Optional[str] = 'openai:deepseek-v3'
    omni_base_url: Optional[str] = ''
    omni_key: Optional[str] = ''
    storage_client: StorageClient = Field(default=create_storage_client())
    simulate_device: Optional[Literal['iPhone 15', 'iPhone 15 Pro', 'iPhone 15 Pro Max', 'iPhone 6'] | str] = None
    debug: Optional[bool] = False

    def copy_and_update(self, **kwargs):
        # TODO: 这里又会实例化一次 StorageClient，建议改成单例模式 @lancefayang
        validated_settings = self.model_validate(kwargs)
        try:
            logger.level('DETAIL', no=20) if validated_settings.debug else logger.level('DETAIL', no=10)
        except ValueError:
            pass
        # 不对 storage_client 进行深拷贝，而是重用原来的实例
        update_dict = validated_settings.model_dump(exclude_none=True)
        if 'storage_client' not in update_dict:
            update_dict['storage_client'] = self.storage_client
        return self.model_copy(update=update_dict, deep=False)


global_settings = Settings()

model_settings = ModelSettings(max_tokens=20000, temperature=0.2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/6/9 15:47
import json
from enum import Enum
from urllib.parse import urlparse, urlencode, quote

import requests
from loguru import logger


class Platform(str, Enum):
    WEB = "WEB"
    WEB_H5 = "WEB_H5"
    QY = "QY"
    KG = "KG"
    KW = "KW"
    BD = "BD"
    MP = "MP"


def get_wx_schema(url):
    """
    :param url: {tenant}://{envVersion}/{path}, eg. musician://release/pages/home/index
    :return:
    """
    scheme, no_schema_url = url.split(':', 1)
    parse_res = urlparse(no_schema_url, scheme=scheme)

    payload = {
        "tenant": parse_res.scheme,
        "path": parse_res.path,
        "query": parse_res.query,
        "envVersion": parse_res.netloc or "release"
    }
    res = requests.post('https://y.tencentmusic.com/passport/v1/wx/generateUrlScheme', json=payload)
    if res.status_code == 200:
        logger.info(res.text)
        schema = res.json().get('data')
        return schema
    else:
        logger.error(res.text)
        return ''


def get_client_url_schema(url, platform):
    """获取客户端打开指定 url schema 地址"""
    if platform == Platform.QY:
        params = {'p': json.dumps({'url': url})}
        return f'qqmusic://qq.com/ui/openUrl?{urlencode(params, quote_via=quote)}'
    elif platform == Platform.KG:
        params = {"cmd": "303", "jsonStr": {"title": "", "url": f"{url}"}, "type": 9, "action": 0}
        return f'kugou://start.weixin?{quote(json.dumps(params))}'
    elif platform == Platform.KW:
        params = {'t': 27, 'u': url}
        return f'kwapp://open?{quote(urlencode(params, quote_via=quote))}'
    elif platform == Platform.BD:
        params = {'t': 27, 'u': url}
        return f'kwapp://open?{quote(urlencode(params, quote_via=quote))}'
    elif platform == Platform.MP:
        return get_wx_schema(url)
    else:
        return url



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/5/30 12:34
from pathlib import Path

from adbutils import AdbDevice
from loguru import logger


class AdbDeviceProxy:
    def __init__(self, device: AdbDevice):
        self._device = device
        self.adb_target_dir = "/data/local/tmp"
        self.adb_path = self._push_yadb()

    def _push_yadb(self):
        if not getattr(self._device, 'yadb', None):
            adb_bin_path = Path(__file__).parent / 'bin' / 'yadb'
            logger.info('push yadb to device...')
            self._device.sync.push(src=adb_bin_path, dst=self.adb_target_dir)
            setattr(self._device, 'yadb', f'{self.adb_target_dir}/yadb')
        return getattr(self._device, 'yadb')

    def execute_command(self, *args: str):
        self._device.shell([
            'app_process',
            f'-Djava.class.path={self.adb_path}',
            f'{self.adb_target_dir}',
            'com.ysbing.yadb.Main',
            *args
        ])

    def input_text(self, text: str):
        self.execute_command('-keyboard', text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author : aidenmo
# @Email : aidenmo@tencent.com
# @Time : 2025/8/3 11:06
from pathlib import Path

from esprima import parseScript
from playwright.async_api import Page, ElementHandle


def parse_script():
    parse_result = {}
    script = (Path(__file__).parent / 'script.js').read_text()
    parsed = parseScript(script, {"range": True})
    for item in parsed.body[0].declarations[0].init.properties:
        parse_result[item.key.name] = script[item.value.range[0]: item.value.range[1]]

    return parse_result


class JSTool:
    _script: dict = parse_script()

    @classmethod
    async def add_highlight_element(cls, page: Page, element_bbox: list) -> ElementHandle:
        return (await page.evaluate_handle(cls._script['add_highlight_element'], [element_bbox])).as_element()

    @classmethod
    async def remove_highlight_element(cls, page: Page):
        return await page.evaluate(cls._script['remove_highlight_element'])

    @classmethod
    async def add_highlight_position(cls, page: Page, x, y) -> ElementHandle:
        return (await page.evaluate_handle(cls._script['add_highlight_position'], [x, y])).as_element()

    @classmethod
    async def remove_highlight_position(cls, page: Page):
        return await page.evaluate(cls._script['remove_highlight_position'])

    @classmethod
    async def has_scrollbar(cls, page: Page, to: str) -> bool:
        if to in ['top', 'bottom']:
            return await page.evaluate(cls._script['has_vertical_scrollbar'])
        else:
            return await page.evaluate(cls._script['has_horizontal_scrollbar'])

    @staticmethod
    async def remove_element(el_handle: ElementHandle):
        await el_handle.evaluate("el => el.remove()")

